@@ -1,5 +1,19 @@
 (in-package :closer-mop)
 
+;; Some utility functions.
+
+(defun required-args (lambda-list &optional (collector #'identity))
+  (loop for arg in lambda-list
+        until (member arg lambda-list-keywords)
+        collect (funcall collector arg)))
+
+(defun ensure-finalized (class &optional (errorp t))
+  (if (typep class 'class)
+    (unless (class-finalized-p class)
+      (finalize-inheritance class))
+    (when errorp (error "~S is not a class." class)))
+  class)
+
 ;; We need a new standard-generic-function for various things.
 
 (cl:defclass standard-generic-function (cl:standard-generic-function)
@@ -115,21 +129,17 @@
 ;; - Removal of direct subclasses.
 
 (cl:defmethod initialize-instance :around
-  ((class standard-class)
-   &rest initargs
-   &key (direct-superclasses ())
-   &allow-other-keys)
+  ((class standard-class) &rest initargs
+   &key (direct-superclasses ()))
   (declare (dynamic-extent initargs))
   (apply #'call-next-method class
-         :optimize-slot-access nil
          :direct-superclasses (modify-superclasses direct-superclasses)
+         :optimize-slot-access nil
          initargs))
 
 (cl:defmethod reinitialize-instance :around
-  ((class standard-class)
-   &rest initargs
-   &key  (direct-superclasses () direct-superclasses-p)
-   &allow-other-keys)
+  ((class standard-class) &rest initargs
+   &key  (direct-superclasses () direct-superclasses-p))
   (declare (dynamic-extent initargs))
   (when direct-superclasses-p
     (setq direct-superclasses (modify-superclasses direct-superclasses))
@@ -138,29 +148,25 @@
           do (remove-direct-subclass superclass class)))
   (if direct-superclasses-p
       (apply #'call-next-method class
-             :optimize-slot-access nil
              :direct-superclasses direct-superclasses
+             :optimize-slot-access nil
              initargs)
     (apply #'call-next-method class
            :optimize-slot-access nil
            initargs)))
 
 (cl:defmethod initialize-instance :around
-  ((class funcallable-standard-class)
-   &rest initargs
-   &key (direct-superclasses ())
-   &allow-other-keys)
+  ((class funcallable-standard-class) &rest initargs
+   &key (direct-superclasses ()))
   (declare (dynamic-extent initargs))
   (apply #'call-next-method class
-         :optimize-slot-access nil
          :direct-superclasses (modify-superclasses direct-superclasses nil)
+         :optimize-slot-access nil
          initargs))
 
 (cl:defmethod reinitialize-instance :around
-  ((class funcallable-standard-class)
-   &rest initargs
-   &key (direct-superclasses () direct-superclasses-p)
-   &allow-other-keys)
+  ((class funcallable-standard-class) &rest initargs
+   &key (direct-superclasses () direct-superclasses-p))
   (declare (dynamic-extent initargs))
   (when direct-superclasses-p
     (setq direct-superclasses (modify-superclasses direct-superclasses nil))
@@ -169,8 +175,8 @@
           do (remove-direct-subclass superclass class)))
   (if direct-superclasses-p
       (apply #'call-next-method class
-             :optimize-slot-access nil
              :direct-superclasses direct-superclasses
+             :optimize-slot-access nil
              initargs)
     (apply #'call-next-method class
            :optimize-slot-access nil
@@ -185,8 +191,7 @@
   ((class forward-referenced-class)
    (new-class funcallable-standard-class)
    &rest initargs
-   &key (direct-superclasses ())
-   &allow-other-keys)
+   &key (direct-superclasses ()))
   (declare (dynamic-extent initargs))
   (apply #'call-next-method class new-class
          :optimize-slot-access nil
@@ -296,9 +301,9 @@
   `(eql ,object))
 
 (cl:defclass eql-specializer* (metaobject)
-  ((object :reader eql-specializer-object
-           :initarg eso
-           :initform (error "Use intern-eql-specializer to create eql-specializers."))
+  ((obj :reader eql-specializer-object
+        :initarg eso
+        :initform (error "Use intern-eql-specializer to create eql-specializers."))
    (direct-methods :reader specializer-direct-methods
                    :accessor es-direct-methods
                    :initform ())))
@@ -416,9 +421,7 @@
                          &key (method-class (generic-function-method-class gf))
                          (qualifiers ())
                          (lambda-list (cadr lambda-expression))
-                         (specializers (loop for arg in lambda-list
-                                             until (member arg lambda-list-keywords)
-                                             collect (find-class 't))))
+                         (specializers (required-args lambda-list (constantly (find-class 't)))))
   (multiple-value-bind
       (method-lambda method-args)
       (make-method-lambda
@@ -477,45 +480,57 @@
                                              (declare ,@declarations)
                                              (block ,name
                                                ,@tail))))
-                  (multiple-value-bind
-                      (method-lambda method-args)
-                      (make-method-lambda
-                       gf (class-prototype method-class)
-                       lambda-expression env)
-                    (with-unique-names (gf method)
-                      (return-from defmethod
-                        `(let ((,gf (if (fboundp ',name)
-                                        (ensure-generic-function ',name)
-                                      (ensure-generic-function
-                                       ',name :lambda-list ',gf-lambda-list)))
-                               (,method
-                                (make-instance
-                                 ',method-class
-                                 :qualifiers ',qualifiers
-                                 :specializers
-                                 (list
-                                  ,@(mapcar
-                                     (lambda (specializer-name)
-                                       (typecase specializer-name
-                                         (symbol `(find-class ',specializer-name))
-                                         (cons (cond
-                                                ((> (length specializer-name) 2)
-                                                 (error "Invalid specializer ~S in defmethod form ~S."
-                                                        specializer-name form))
-                                                ((eq (car specializer-name) 'eql)
-                                                 `(intern-eql-specializer ,(cadr specializer-name)))
-                                                (t (error "Invalid specializer ~S in defmethod form ~S."
-                                                          specializer-name form))))
-                                         (t (error "Invalid specializer ~S in defmethod form ~S."
-                                                   specializer-name form))))
-                                     (extract-specializer-names specialized-args)))
-                                 :lambda-list ',lambda-list
-                                 :function (function ,method-lambda)
-                                 ,@(unless (eq documentation :unbound)
-                                     (list :documentation documentation))
-                                 ,@method-args)))
-                           (add-method ,gf ,method)
-                           ,method)))))))))
+                  (if (equal (compute-applicable-methods
+                              #'make-method-lambda
+                              (list gf (class-prototype method-class)
+                                    lambda-expression env))
+                             (list (find-method
+                                    #'make-method-lambda '()
+                                    (list (find-class 'cl:standard-generic-function)
+                                          (find-class 'standard-method)
+                                          (find-class 't)
+                                          (find-class 't))
+                                    nil)))
+                      (return-from defmethod `(cl:defmethod ,@(rest form)))
+                    (multiple-value-bind
+                        (method-lambda method-args)
+                        (make-method-lambda
+                         gf (class-prototype method-class)
+                         lambda-expression env)
+                      (with-unique-names (gf method)
+                        (return-from defmethod
+                          `(let ((,gf (if (fboundp ',name)
+                                          (ensure-generic-function ',name)
+                                        (ensure-generic-function
+                                         ',name :lambda-list ',gf-lambda-list)))
+                                 (,method
+                                  (make-instance
+                                   ',method-class
+                                   :qualifiers ',qualifiers
+                                   :specializers
+                                   (list
+                                    ,@(mapcar
+                                       (lambda (specializer-name)
+                                         (typecase specializer-name
+                                           (symbol `(find-class ',specializer-name))
+                                           (cons (cond
+                                                  ((> (length specializer-name) 2)
+                                                   (error "Invalid specializer ~S in defmethod form ~S."
+                                                          specializer-name form))
+                                                  ((eq (car specializer-name) 'eql)
+                                                   `(intern-eql-specializer ,(cadr specializer-name)))
+                                                  (t (error "Invalid specializer ~S in defmethod form ~S."
+                                                            specializer-name form))))
+                                           (t (error "Invalid specializer ~S in defmethod form ~S."
+                                                     specializer-name form))))
+                                       (extract-specializer-names specialized-args)))
+                                   :lambda-list ',lambda-list
+                                   :function (function ,method-lambda)
+                                   ,@(unless (eq documentation :unbound)
+                                       (list :documentation documentation))
+                                   ,@method-args)))
+                             (add-method ,gf ,method)
+                             ,method))))))))))
 
 ;; The following macro ensures that the new standard-generic-function
 ;; is used by default. It also ensures that make-method-lambda is called
